// qlcp-panel.js
'use strict';

//****named constants****************************************************************************************************************************************************************
const CPF_VERSION = 5;

const SERVER_PORT_DEFAULT = 9999;                         // Default port for QLCP-NET server
const NUM_INDICATION_BITS_DEFAULT = 512;                  // Default number of indication bits

const OPACITY_KNOWN = 1.0;                                // Component opacity level for 'known' state
const OPACITY_UNKNOWN = 0.3;                              // Component opacity level for 'unknown' (or 'uncertain') state

//****global variables***************************************************************************************************************************************************************
let redrawComponents = false;           // Indicates component layer needs to be drawn/updated (occurs at a browser determined interval in requestAnimationFrame processing.)
let controlOptionsStr = "";             // Control label/bit name string used to set the innerHTML of properties select options.
let indicationOptionsStr = "";          // Indication label/bit name string used to set the innerHTML of properties select options.
let nbrRows = 60;                       // Max rows in editor grid.
let nbrCols = 182;                      // Max columns in editor grid.
let numIndicationBits = NUM_INDICATION_BITS_DEFAULT;  // Number of indication bits
let controlBits = new Array(512);       // Stores control bit states for latched controls (e.g. toggle switch, keyswitch, rotary switch ,etc.
let hostUrl = window.location.host;     // IP address of the QLCP-NET server that is entered for the URL in the browser
let serverPort = SERVER_PORT_DEFAULT;   // IP port to use when accessing the QLCP-NET server
let passcode = "";                      // Passcode for locked controls.
let lastClickTime = new Date();         // Time when user last clicked on a control component (for passcode functionality.)
let initLatchedSwitches = false;        // Set true when changing between run and design mode to initialize latched switches
let selectedComponents = [];            // Contains the currently selected components.
let mouseDownOnComponent = false;       // Set by a component's 'mousedown' event, cleared on stage 'mouseup' event.
let selectionRect = null;               // Selection rectangle object displayed while dragging to select.
let selStart = { x: 0, y:0 };           // Starting coordinates for a mouse drage selection.
let allow_bgImg_selection = false;
let lastTextAlign = 'left';
let ctrlKeyDown = false;                // Capture when control key is pressed
let idNbr = 0;                          // The next id number to assign a component.

const defaultFillColor = 'rgba(85,85,85,1)';

// Key:Value pairs that track drop modes for each component type
// The boolean value is true when the drop mode for that component type is active.
let dropMode = {
  text: false,
  LED: false,
  pb: false,
  pb_square: false,
  togSw: false,
  togSwMom: false,
  quadrilateral: false,
  arrow: false,
  timer: false,
  keyswitch: false,
  rotaryswitch: false,
  none: true
};

// Colors used for drawing components and text, and assigning indications.
const stdColors = ["red", "yellow", "lime",
                  "green", "orange", "blue",
                  "white", "black", "cyan",
                  "dodgerblue", "gray", "dimgray",
                  "transparent"];
const colorVals =  ["rgba(255,0,0,1)", "rgba(255,255,0,1)", "rgba(0,255,0,1)",
                  "rgba(0,128,0,1)", "rgba(255,165,0,1)", "rgba(0,0,255,1)",
                  "rgba(255,255,255,1)", "rgba(0,0,0,1)", "rgba(0,255,255,1)",
                  "rgba(30,144,255,1)", "rgba(128,128,128,1)", "rgba(105,105,105,1)",
                  "rgba(0,0,0,0)"];

/******************************************************************************/
/******************************************************************************/
//Create the stage for component and text.
//44 pixels is based on the grid size, overlap of grids, etc.
let stage = new Konva.Stage({
  container: 'component_container',
  width: nbrCols * 44,
  height: nbrRows * 44
});

//create the component layer and add it to the stage
let component_layer = new Konva.Layer();
let bgImage_layer = new Konva.Layer();
stage.add(bgImage_layer, component_layer);

let scrollContainer = document.getElementById('scroll_container');
scrollContainer.addEventListener('scroll', function () {
  let dx = scrollContainer.scrollLeft;
  let dy = scrollContainer.scrollTop;
  stage.container().style.transform = 'translate(' + dx + 'px, ' + dy + 'px)';
  stage.x(-dx);
  stage.y(-dy);
  stage.batchDraw();
});

// Add a selection box rectangle to the stage.
CreateSelectionRect();
/******************************************************************************/
/******************************************************************************/


function RGBAToHexA(rgba) {
  let ex = /^rgba\((((((((1?[1-9]?\d)|10\d|(2[0-4]\d)|25[0-5]),\s?)){3})|(((([1-9]?\d(\.\d+)?)|100|(\.\d+))%,\s?){3}))|(((((1?[1-9]?\d)|10\d|(2[0-4]\d)|25[0-5])\s){3})|(((([1-9]?\d(\.\d+)?)|100|(\.\d+))%\s){3}))\/\s)((0?\.\d+)|[01]|(([1-9]?\d(\.\d+)?)|100|(\.\d+))%)\)$/i;
  if (ex.test(rgba)) {
    let sep = rgba.indexOf(",") > -1 ? "," : " ";
    rgba = rgba.substr(5).split(")")[0].split(sep);

    // strip the slash if using space-separated syntax
    if (rgba.indexOf("/") > -1)
      rgba.splice(3,1);

    for (let R in rgba) {
      let r = rgba[R];
      if (r.indexOf("%") > -1) {
        let p = r.substr(0,r.length - 1) / 100;

        if (R < 3) {
          rgba[R] = Math.round(p * 255);
        } else {
          rgba[R] = p;
        }
      }
    }

    let r = (+rgba[0]).toString(16),
      g = (+rgba[1]).toString(16),
      b = (+rgba[2]).toString(16),
      a = Math.round(+rgba[3] * 255).toString(16);

    if (r.length == 1)
      r = "0" + r;
    if (g.length == 1)
      g = "0" + g;
    if (b.length == 1)
      b = "0" + b;
    if (a.length == 1)
      a = "0" + a;

    return "#" + r + g + b + a;

  } else {
    return "Invalid input color";
  }
} // end RGBAToHex()

function hexAToRGBA(h,isPct) {
  let ex = /^#([\da-f]{4}){1,2}$/i;
  if (ex.test(h)) {
    let r = 0, g = 0, b = 0, a = 1;
    isPct = isPct === true;

    if (h.length == 5) {
      r = "0x" + h[1] + h[1];
      g = "0x" + h[2] + h[2];
      b = "0x" + h[3] + h[3];
      a = "0x" + h[4] + h[4];

    } else if (h.length == 9) {
      r = "0x" + h[1] + h[2];
      g = "0x" + h[3] + h[4];
      b = "0x" + h[5] + h[6];
      a = "0x" + h[7] + h[8];
    }
    a = +(a / 255).toFixed(3);
    if (isPct) {
      r = +(r / 255 * 100).toFixed(1);
      g = +(g / 255 * 100).toFixed(1);
      b = +(b / 255 * 100).toFixed(1);
      a = +(a * 100).toFixed(1);
    }

    return "rgba("+ (isPct ? r + "%," + g + "%," + b + "%" + "," + a : +r + "," + +g + "," + +b + "," + a) + ")";

  } else {
    return "Invalid input color";
  }
} // end hexAToRGBA()

function CreateSelectionRect() {
// Create a selection rewctangle to be drawn when selecting components.
  selectionRect = new Konva.Rect({
    x: 0,
    y: 0,
    height: 0,
    width: 0,
    fill: "rgba(0,0,0,0)",
    // dash: [2,2],
    stroke: "rgba(255,255,255,1)",
    strokeWidth: 1,
    visible: false,
    listening: false
  });
  component_layer.add(selectionRect);
} // end CreateSelectionRect()

function initSelectionRect(posIn) {
  selectionRect.setAttrs({
    height: 0,
    width: 0,
    visible: true
  });
} // initSelectionRect()

function componentSelected(cmp, set) {
  cmp.attrs.selected = set;
  if (cmp.attrs.selected) {
    cmp.find('.selection')[0].attrs.show();
    // Add to the selectedComponents array.
    selectedComponents.push(cmp);
  } else {
    cmp.find('.selection')[0].attrs.hide();
    // Remove from the selectedComponents array.
    let idx = selectedComponents.indexOf(cmp);
    if (idx > -1) {
      selectedComponents.splice(idx,1);
    }
  }
} // end componentSelected()

function drawSelectionRect(posIn) {

  selectionRect.setAttrs({
    x: Math.min(selStart.x, posIn.x),
    y: Math.min(selStart.y, posIn.y),
    width: Math.abs(posIn.x - selStart.x),
    height: Math.abs(posIn.y - selStart.y)
  });

  let box = selectionRect.getClientRect();

  // Run the collision check.
  for(const comp of component_layer.getChildren()) {
    if (comp !== selectionRect)
    {
      // Modify the components selected state if it has changed.
      if (comp.attrs.selected != Konva.Util.haveIntersection(box,comp.getClientRect())) {
        componentSelected(comp, !comp.attrs.selected);

        // Show the position if a single component is selected.
        if (selectedComponents.length === 1) {
          if ($( "#properties_dialog" ).dialog( "isOpen" )) {
            ComponentProperties(selectedComponents[0]);
          }
        } else {
          // Clear properties dialog box
          document.getElementById("properties_dialog").innerHTML = "";
        }
      };
    }
  }
} // drawSelectionRect()

function ClearSelectedComponents() {
  for (const comp of selectedComponents) {
    comp.find('.selection')[0].attrs.hide();
    comp.attrs.selected = false;
  }
  selectedComponents.length = 0;

  // Clear properties dialog box
  document.getElementById("properties_dialog").innerHTML = "";
} // end ClearSelectedComponents()

function ComponentMouseDown(comp) {
  mouseDownOnComponent = true;

  // Toggle the component selected state if control key was pressed.
  if (ctrlKeyDown) {
    componentSelected(comp, !comp.attrs.selected);
  } else {
    if (!selectedComponents.includes(comp)) {
      ClearSelectedComponents();
      componentSelected(comp,true);
    }
  }
} // end ComponentMouseDown()

// The following three mouse event routines are for selecting and moving components on the
// stage (component layer). The reason the event listeners are bound to the stage is so
// they only fire when the event occurs in the stage area, not the title, menu, or footer areas.
// This event depends on bubbling of events from the component elements up to the component layer.
// It depends on the component setting the mouseDownOnComponent flag first.
stage.getContent().addEventListener("mousedown", function (e) {

  selStart = { x: e.layerX, y: e.layerY };

  if (mouseDownOnComponent) {
    allow_bgImg_selection = false;
  } else { // If the mouse down event was not on a component.
    allow_bgImg_selection = true;

    // Clear any selected components.
    ClearSelectedComponents();

    // If there is a drop mode active, add the component to the stage, otherwise,
    // prepare to draw a selection rectangle.
    if (dropMode['none']) {
      // Start formation of selection rectangle unless background image drag is active.
      if (!(bgImage_layer.hasChildren() && bgImage_layer.getChildren()[0].draggable())) {
        initSelectionRect();
      }
    } else {
      // AddComponent function will check drop mode to determine which component to add.
      let comp = AddComponent(Math.round(e.layerX), Math.round(e.layerY), ++idNbr);
      componentSelected(comp, true);
      mouseDownOnComponent = true;
    }
  }

  // If there is a single component selected, show it's properties.
  if (selectedComponents.length === 1) {
    if ($( "#properties_dialog" ).dialog( "isOpen" )) {
      ComponentProperties(selectedComponents[0]);
    }
  } else {
    // Clear properties dialog box
    document.getElementById("properties_dialog").innerHTML = "";
  }
}); // end stage mousedown

stage.getContent().addEventListener('mousemove', function (e) {
  // This function depends on requestAnimationFrame to redraw the component layer.
  // If the selection rectangle is visible, just keep updating it.
  if (selectionRect.visible()) {
    drawSelectionRect({x: e.layerX, y: e.layerY});
  } else {
    if (mouseDownOnComponent) {
      // Move the currently selected components if the position lock is false.
      if (!TogglePositionLock.positionLocked) {

        // Determine the delta x and y values.
        let deltaX = e.layerX - selStart.x;
        let deltaY = e.layerY - selStart.y;

        // Move all selected components.
        for (const comp of selectedComponents) {
          comp.position({
            x: comp.position().x + deltaX,
            y: comp.position().y + deltaY
          });
        }

        // Update the starting coordinates for the next move event.
        selStart.x = e.layerX;
        selStart.y = e.layerY;
      }

      // Show the currently selected component's (or image) x,y position.
      if (bgImage_layer.hasChildren() && bgImage_layer.getChildren()[0].draggable()) {

      } else if (selectedComponents.length == 1) {
        if ($( "#properties_dialog" ).dialog( "isOpen" )) {
          document.getElementById("x-pos-value").value = selectedComponents[0].attrs.x;
          document.getElementById("y-pos-value").value = selectedComponents[0].attrs.y;
        }
      }
    }
  }
}); // end stage mousemove

//This function depends on requestAnimationFrame to redraw the component layer.
stage.getContent().addEventListener("mouseup", function ()
{
  // Clear flags for next selection
  mouseDownOnComponent = false;
  selectionRect.visible(false);
}); // end stage.event('mouseup')

/******************************************************************************/
/******************************************************************************/


//This function is periodically called using requestAnimationFrame to draw the component
//layer. This is required when changes are made to the appearance of a component.
//Since design mode will typically be used on a fast enough laptop PC, the component layer
//is drawn as frequently as the browser can. For run mode, the redrawComponents flag is used
//to only draw the layer when necessary.
function UpdateComponentLayer() {

  component_layer.draw();

  //requestAnimationFrame works using recursion
  window.requestAnimationFrame(UpdateComponentLayer);
} // end UpdateComponentLayer()

function InitGrid() {
  //Create the divs for the grid lines
  let htmlGridLines = "";

  for (let rowIndex=0; rowIndex < nbrRows; rowIndex++)
  {
    for(let colIndex=0; colIndex < nbrCols; colIndex++)
    {
      htmlGridLines += '<div class="gridLines" id="s' + rowIndex + ',' + colIndex + '" style="float: left"></div>';
    }
    // start the next row (no id assigned to cell, just a fixed block got to next row)
    htmlGridLines += '<div class="gridLines"></div>';
  }

  document.getElementById("grid_lines").innerHTML = htmlGridLines;

} // end InitGrid()

function InitDialogBoxes() {
  $( "#properties_dialog" ).dialog({ autoOpen: false });
  $( "#properties_dialog" ).dialog({ minWidth: "510" });
  $( "#properties_dialog" ).dialog({ height: "700" });
  $( "#properties_dialog" ).dialog({ resizable: true });
  $( "#properties_dialog" ).dialog({ appendTo: "#buttonBackground" });

  $( "#message_dialog" ).dialog({ autoOpen: false });
  $( "#message_dialog" ).dialog({ width: "auto" });
  $( "#message_dialog" ).dialog({ height: "auto" });
  $( "#message_dialog" ).dialog({ resizable: false });
  $( "#message_dialog" ).dialog({ modal: true });
  $( "#message_dialog" ).dialog({ appendTo: "#buttonBackground" });
} // end InitDialogBoxes()
function ShowMessageDialog() {
  //This function is called to show a message dialog. The innerHTML and title of the dialog
  //are set by the caller
  $( "#message_dialog" ).dialog( "option", "buttons", [{
      text: "Close", click: function() {
        $( this ).dialog( "close" );
      }
    }
  ]);

  $( "#message_dialog" ).dialog( "open" );
} // end ShowMessageDialog()

function SecondsElapsed(startTime) {
  let currentTime = new Date();
  let timeDiff = currentTime - startTime; //in ms

  // strip the ms
  timeDiff /= 1000;

  // return seconds
  return Math.round(timeDiff);
} // end SecondsElapsed()

function MillisecondsElapsed(startTime) {
  let currentTime = new Date();
  return currentTime - startTime;
} // end MillisecondsElapsed()

//This function is called when the user clicks on the Open file button
function RequestPanelFileDialog() {

    ////////////// Open a file and read contents. ////////////////
    var fileInput = document.createElement("input");
    fileInput.type = "file";
    fileInput.accept = '.cpf';

    fileInput.onchange = e => {

      // Get a hold of the file reference.
      var file = e.target.files[0];

      if (file.name.match(/\.(cpf|)$/)) {
        // Set up the reader.
        var reader = new FileReader();

        reader.readAsText(file, 'UTF-8');

      } else {
        alert("File not supported, .cpf files only.");
      }

      reader.onload = readerEvent => {
        var content = readerEvent.target.result;
        OpenPanel(content);
      }
    };

    fileInput.click();
    ///////////////// file open ////////////////

} // end RequestPanelFileDialog()

function OpenPanel(data) {

  let bgImgSz = {};
  let bgImgPos = {};
  let bgImgObj = {};
  let bgImgSrc = undefined;

  // Clear any previous selection.
  selectedComponents.length = 0;

  //clear the component layer
  component_layer.destroyChildren();
  component_layer.clear();

  CreateSelectionRect();

  //remove any background image for the panel (component style designs)
  bgImage_layer.removeChildren();
  bgImage_layer.clear();

  InitGrid();

  // Store each line from the data in an array
  let attrLines = new Array();
  if (data.indexOf("\r") > 0) {
    attrLines = data.split("\r\n");
  } else if (data.indexOf("\n") > 0) {
    attrLines = data.split("\n");
  }

  // Reset id number
  idNbr = 1;

  // Get the version from the first line.
  let compAttr = attrLines.shift().split("=");

  if (+compAttr[1] > CPF_VERSION) {
    $( "#message_dialog" ).dialog( "option", "title", "Control Panel Request" );

    document.getElementById("message_dialog").innerHTML = "<div><img style='float: left; margin-top: 5px; width: 36px' src='images/warning_icon.png'/>"
      + "<div style='float: left; margin-left: 15px; margin-top: 6px; width: 350px'>The panel file being loaded was created with a newer "
      + "version of the QLCP Web GUI SW. Please update the QLCP's Web GUI SW to view this control panel.<br/><br/>"
      + "</div></div>";

    ShowMessageDialog();
    return;
  }

  // Parse each line the panel file.
  while (attrLines.length > 0) {

    // Get the name/value pair (the attribute on this line)
    compAttr = attrLines.shift().split(":");

    // Handle the entry based on the type of object/component it defines.
    switch (compAttr[0]) {

      case "icon_begin":
      {
        alert("Mosaic Tile panels are no longer supported. Please use web-client version 4.11 or earlier..");
        return;
      }
      break;

      // Load an LCP background image.
      case "lcp_background":
      {
        // Create the image object and save the data for use after all params are read.
        bgImgObj = new Image();
        bgImgSrc = compAttr[1] + ":" + compAttr[2];

        bgImgObj.onload = () => {
          let bgImg = new Konva.Image({
            image: bgImgObj,
            x: 20, // default position from previously handled with CSS margin values.
            y: 50,
            stroke: 'rgba(0,255,255,1)',
            strokeWidth: 3,
            strokeEnabled: false,
            draggable: false,
            id: "c" + idNbr++ // todo...not yet sure how image fits with other components
          });

          let actual_sz = bgImg.size();

          if (bgImgSz.width != undefined) {
            // Maintain original aspect ratio to determine height.
            bgImg.size({
              width: bgImgSz.width,
              height: (bgImgSz.width * (actual_sz.height/actual_sz.width))
            });
          }

          if (bgImgPos != undefined) {
            bgImg.position({
              x: bgImgPos.x,
              y: bgImgPos.y
            });
          }

          bgImage_layer.add(bgImg);
          bgImage_layer.draw();
        }
      }
      break;

      // Store an LCP background image width.
      case "lcp_background_width":
      {
          bgImgSz.width = Number(compAttr[1]);
      }
      break;

      // Store an LCP background image position.
      case "lcp_background_pos":
      {
        bgImgPos.x = Number(compAttr[1].split(',')[0]);
        bgImgPos.y = Number(compAttr[1].split(',')[1]);
      }
      break;

      // Load a single component: read all component attributes, create the component and add it to the layer.
      // Note: "text_begin" and "text_end" is for older panel files.
      case "component_begin":
      case "text_begin":
      {
        let attributes = new Array();
        let indicationAttrs = new Array();
        let controlAttrs = new Array();

        // Split the next line name/value pair
        compAttr = attrLines.shift().split(":");

        // Proccess all attributes until the end of component is detected.
        while ((compAttr[0] !== "component_end") && (compAttr[0] !== "text_end")) {

          let attrName = compAttr[0];
          let attrValue = compAttr[1];

          // Adjust any color values to use new rgba internal representation instead
          // of HTML color strings that were used on older versions.
          if ((attrName == 'fill' )|| (attrName == 'stroke') || (attrName == 'defaultColor')) {

            // Attempt to look up the value by the color name (for older saved panels.)
            let idx = stdColors.indexOf(attrValue);
            if (idx != -1) {
              attrValue = colorVals[idx];
            }
          }

          // Only store height and width if not flagged as label section.
          if (attrName === 'width' || attrName === 'height') {
            if (attributes.name !== "label") {
              // Store in the attributes array as a name/value pair.
              attributes[attrName] = attrValue;
            }
          } else {
            // Store in the attributes array as a name/value pair.
            attributes[attrName] = attrValue;
          }

          // Search for control attributes
          if (attrName.indexOf('control') >= 0) {
            controlAttrs.push(attrValue);
          }

          // Search for indication attributes
          if (attrName.indexOf('indication') >= 0) {

            // Revert to default color assignments if no color was stored.
            if (attrValue.indexOf('rgba') === -1) {
              // Add a value of true if there was no name assignment to make the string complete.
              if (attrValue.split(",")[0] == '0') {
                attrValue += ',0';
              }
              attrValue += ',' + colorVals[attrName.replace('indication', '') % (colorVals.length)];
            }
            indicationAttrs.push(attrValue);
          }

          // Get the next line of panel file.
          compAttr = attrLines.shift().split(":");
        } // end while ... (process all attributes)

        // Set the component type to text if it is undefined.
        if (attributes.type === undefined) {
          attributes.type = "text"
        }

        // Create a new component object from the stored attributes.
        dropMode[attributes.type] = true;
        let Component = AddComponent(Math.round(Number(attributes.x)), Math.round(Number(attributes.y)), idNbr);
        dropMode[attributes.type] = false;

        if (Component == null) {
          alert("Error! Unable to create component. Panel file will not be loaded.");
          return;
        }

        // Add the id number to the component attributes and increment the count.
        attributes["id"] = "c" + idNbr++;

        SetComponentProperties(Component, attributes);

        // Set the control properties for this component
        for (let i = 0; i < controlAttrs.length; i++) {
          Component.attrs["control" + i] = controlAttrs[i];
        }

        // Set the indication properties for this component
        for (let i = 0; i < indicationAttrs.length; i++) {
          Component.attrs["indication" + i] = indicationAttrs[i];
        }

        // Update the selection border (needed for older panels that were saved before selection border was updated)
        if (typeof Component.find('Text')[0] !== 'undefined') {
          Component.find('Text')[0].setText(Component.find('Text')[0].attrs.text);
        }
        Component.find('.selection')[0].attrs.update();
      }
      break;
    } // end switch(compAttr[0])
  } // end while (attrLines.length > 0)

  // Now that file parsing is done, load/handle background image.
  if (bgImgSrc != undefined) {
    bgImgObj.src = bgImgSrc;
  }

  redrawComponents = true;

} // end OpenPanel()

function SetComponentProperties(Component, attributes) {
  let indicatorObj;
  let textObj;
  let cmpType = Component.attrs.type;

  if (Component.find('.indicator').length > 0) {
    indicatorObj = Component.find('.indicator')[0];
  }

  if (Component.find('Text').length > 0) {
    textObj = Component.find('Text')[0];
  }

  // Set pb_square text properties.
  if (cmpType === "pb_square") {
    if (typeof attributes.text !== 'undefined') {
      textObj.setText(attributes.text);
    }
    if (typeof attributes.fill !== 'undefined') {
      textObj.fill(attributes.fill);
    }
    if (typeof attributes.fontSize !== 'undefined') {
      textObj.fontSize(+attributes.fontSize);
    }

    let base = Component.find('Rect')[0];
    let selectRect = Component.find('Rect')[2];

    if (attributes.width !== undefined) {
      indicatorObj.width(+attributes.width);
      base.width(+attributes.width+10);
    }
    if (attributes.height !== undefined) {
      indicatorObj.height(+attributes.height);
      base.height(+attributes.height+10);
    }

    selectRect.width(base.width());
    selectRect.height(base.height());

    // Align text before scaling!
    textObj.attrs.alignText();
  }

  // Set text or timer text properties.
  if (cmpType === "text" || cmpType === "timer") {

    if (typeof attributes.defaultColor !== 'undefined') {
      indicatorObj.attrs.defaultColor = attributes.defaultColor;
    } else {
      // For older versions that don't support default color.
      indicatorObj.attrs.defaultColor = "rgba(255,255,255,1)";
    }

    if (typeof attributes.fill !== 'undefined') {
      indicatorObj.attrs.fill = attributes.fill;
    }

    if (typeof attributes.fontSize !== 'undefined') {
      textObj.fontSize(+attributes.fontSize);
    }

    if (typeof attributes.text !== 'undefined') {
      textObj.setText(attributes.text);
    }

    if(cmpType === "text") {
      if (typeof attributes.align !== 'undefined') {
        textObj.align(attributes.align);
        textObj.attrs.alignText();
      }
    }

  } else if (typeof attributes.scaleX !== 'undefined') {
    //scale applies to all components but text,
    //and yes, scaleX is used for y.
    Component.scale({
      x: +attributes.scaleX,
      y: +attributes.scaleX
    });
  }

  if (cmpType === "timer") {
    indicatorObj.attrs.timerDuration = +attributes.timerDuration;
  }

  // Set LED, quadrilateral, or arrow stroke values.
  if (cmpType === "LED" || cmpType === "quadrilateral" || cmpType === "arrow") {

    if (typeof attributes.stroke !== 'undefined') {
      indicatorObj.stroke(attributes.stroke);
    }

    if (typeof attributes.strokeWidth !== 'undefined') {
      indicatorObj.strokeWidth(+attributes.strokeWidth);
      indicatorObj.strokeEnabled((+attributes.strokeWidth > 0 ? true: false));
    }

    if (typeof attributes.defaultColor !== 'undefined') {
      indicatorObj.attrs.defaultColor = attributes.defaultColor;
    }

    if (typeof attributes.fill !== 'undefined') {
      indicatorObj.fill(attributes.fill);
    }
  }

  // Set components with rotation property.
  if (cmpType === "text" || cmpType === "quadrilateral" || cmpType === "arrow" ||
      cmpType === "timer" || cmpType === "togSw" || cmpType === "togSwMom") {

    if (typeof attributes.rotation !== 'undefined') {
      Component.rotation(+attributes.rotation);
    }
  }

  // Set quadrilateral point coordinates.
  if (cmpType === "quadrilateral") {

    let quad = indicatorObj.attrs;
    quad.point1x = +attributes.point1x;
    quad.point1y = +attributes.point1y;
    quad.point2x = +attributes.point2x;
    quad.point2y = +attributes.point2y;
    quad.point3x = +attributes.point3x;
    quad.point3y = +attributes.point3y;
    quad.point4x = +attributes.point4x;
    quad.point4y = +attributes.point4y;
  }

  if (Component.attrs.nbrCtrls > 0) {

    if (typeof attributes.passcode !== 'undefined') {
      Component.attrs.passcode = attributes.passcode;
    }

    if (typeof attributes.timeout !== 'undefined') {
      Component.attrs.timeout = +attributes.timeout;
    }
  }

  if (Component.attrs.hasIndications) {
    if (typeof attributes.flashTime !== 'undefined') {
      Component.attrs.flashTime = +attributes.flashTime;
    }
  }
} // end SetComponentProperties()

//********************* Component functions/classes ****************************
//******************************************************************************
// All components are groups of shapes that inherit the common properties and
// methods of the group created in this function.
function ComponentGroup(xpos, ypos, id_nbr, cmp_type, cmp_text, numCtrls, has_indications) {
  let Component = new Konva.Group({
    x: xpos,
    y: ypos,
    scaleX: 1,
    scaleY: 1,
    offsetX: 0,
    offsetY: 0,
    skewX: 0,
    skewY: 0,
    type: cmp_type,
    text: cmp_text,
    nbrCtrls: numCtrls,
    hasIndications: has_indications,
    state: false,
    rotation: 0,
    flashTime: 500,
    flashTimeCntr: 0,
    flashState: 1,
    passcode: '',
    unlocked: false,
    timeout: 0,
    selected: false,
    draggable: (TogglePositionLock.positionLocked == true ? false : true),
    releaseTimeout: 0,
    id: "c" + id_nbr
  });

  return Component;
} // end ComponentGroup()

//All components inherit the selection border rectangle and sizing handles in this function
function ComponentSelectionBorder() {
  let selectionBorder = new Konva.Rect({
    name: "selection",
    fill: "rgba(0,0,0,0)",
    stroke: "rgba(0,255,255,1)",
    strokeWidth: 2,
    strokeEnabled: false,

    update: function()
    {
      if (selectionBorder.getParent().attrs.type === 'quadrilateral' ||
          selectionBorder.getParent().attrs.type === 'LED' ||
          selectionBorder.getParent().attrs.type === 'text' ||
          selectionBorder.getParent().attrs.type === 'timer')
      {
        let component = selectionBorder.getParent();
        let borderValues = component.find('.indicator')[0].attrs.getBorderValues();
        selectionBorder.position({
          x: borderValues.x,
          y: borderValues.y
        });
        selectionBorder.attrs.width = borderValues.width;
        selectionBorder.attrs.height = borderValues.height;
      }
    },

    show: function()
    {
      selectionBorder.strokeEnabled(true);
    },

    hide: function()
    {
      selectionBorder.strokeEnabled(false);
    }
  });

  return selectionBorder;
} // end ComponentSelectionBorder()

//This function creates a new quadrilateral.
function CreateQuadrilateral(xpos, ypos, id_nbr) {
  //create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'quadrilateral', '', 0, true);

  //create a quadrilateral shape (rectangle initially)
  let quad = new Konva.Shape({
    name: 'indicator', //this is the shape that shows the indications for the component
    fill: defaultFillColor,
    defaultColor: defaultFillColor,
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 0,
    strokeEnabled: false,
    point1x: 0,
    point1y: 0,
    point2x: 70,
    point2y: 0,
    point3x: 70,
    point3y: 18,
    point4x: 0,
    point4y: 18,

    // a Konva.Canvas renderer is passed into the sceneFunc function
    sceneFunc: function(context, shape)
    {
      context.beginPath();
      context.moveTo(this.attrs.point1x, this.attrs.point1y);
      context.lineTo(this.attrs.point2x, this.attrs.point2y);
      context.lineTo(this.attrs.point3x, this.attrs.point3y);
      context.lineTo(this.attrs.point4x, this.attrs.point4y);
      context.closePath();
      context.fillStrokeShape(shape);
    },

    getBorderValues: function()
    {
      let x, y, width, height;

      x = -quad.attrs.strokeWidth/Component.attrs.scaleX/2;
      y = -quad.attrs.strokeWidth/Component.attrs.scaleX/2;

      //for width, use the largest x value
      if (quad.attrs.point2x > quad.attrs.point3x)
      {
        width = Number(quad.attrs.point2x) + (quad.attrs.strokeWidth/Component.attrs.scaleX);
      }
      else
      {
        width = Number(quad.attrs.point3x) + (quad.attrs.strokeWidth/Component.attrs.scaleX);
      }

      //height, use the largest y value
      if (quad.attrs.point3y > quad.attrs.point4y)
      {
        height = Number(quad.attrs.point3y) + (quad.attrs.strokeWidth/Component.attrs.scaleX);
      }
      else
      {
        height = Number(quad.attrs.point4y) + (quad.attrs.strokeWidth/Component.attrs.scaleX);
      }

      return {
        x: x,
        y: y,
        width: width,
        height: height
      };
    }
  });

  //we don't want the stroke width to scale
  quad.strokeScaleEnabled(false);

  //create a selection border for this component and set its initial width and height
  let selectionBorder = ComponentSelectionBorder();
  let borderValues = quad.attrs.getBorderValues();
  selectionBorder.position({
    x: borderValues.x,
    y: borderValues.y
  });
  selectionBorder.attrs.width = borderValues.width;
  selectionBorder.attrs.height = borderValues.height;
  selectionBorder.strokeScaleEnabled(false);

  //add shapes to the group
  Component.add(quad, selectionBorder);

  //mouse down event
  Component.on('mousedown', function() {
    ComponentMouseDown(this);
  });

  return Component;
} // end CreateQuadrilateral()

//This function creates a new arrow.
function CreateArrow(xpos, ypos, id_nbr) {
  //create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'arrow', '', 0, true);

  //create an arrow
  let arrow = new Konva.Shape({
    name: "indicator",
    fill: defaultFillColor,
    defaultColor: defaultFillColor,
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 0,
    strokeEnabled: false,

    // a Konva.Canvas renderer is passed into the sceneFunc function
    sceneFunc: function(context, shape)
    {
      context.beginPath();
      context.moveTo(0,0);
      context.lineTo(20,0);
      context.lineTo(20,-5);
      context.lineTo(40,5);
      context.lineTo(20,15);
      context.lineTo(20,10);
      context.lineTo(0,10);
      context.closePath();
      context.fillStrokeShape(shape);
    }
  });

  arrow.strokeScaleEnabled(false);

  //create a highlight for this component that inherits the common component highlight properties
  let selectionBorder = ComponentSelectionBorder();
  selectionBorder.attrs.y = -5;
  selectionBorder.attrs.width = 40;
  selectionBorder.attrs.height = 20;
  selectionBorder.strokeScaleEnabled(false);

  //add shapes to the group
  Component.add(arrow, selectionBorder);

  //mouse down event
  Component.on('mousedown', function() {
    ComponentMouseDown(this);
  });

  return Component;
} // end CreateArrow()

//This function creates a new LED.
function CreateLED(xpos, ypos, id_nbr) {
  //create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'LED', '', 0, true);

  //create the led shape
  let led = new Konva.Circle({
    name: "indicator",
    radius: 12,
    fill: defaultFillColor,
    defaultColor: defaultFillColor,
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 2,
    strokeEnabled: true,

    getBorderValues: function()
    {
      let x, y, width, height;

      x = -(Number(led.attrs.radius) + led.attrs.strokeWidth/Component.attrs.scaleX/2);
      y = -(Number(led.attrs.radius) + led.attrs.strokeWidth/Component.attrs.scaleX/2);
      width = 2 * Number(led.attrs.radius) + led.attrs.strokeWidth/Component.attrs.scaleX;
      height = width;

      return {
        x: x,
        y: y,
        width: width,
        height: height
      };
    }
  });

  led.strokeScaleEnabled(false);
  //led.perfectDrawEnabled(false); //this could help performance

  //create a selection border for this component and set its initial position, width, and height
  let selectionBorder = ComponentSelectionBorder();
  let borderValues = led.attrs.getBorderValues();
  selectionBorder.position({
    x: borderValues.x,
    y: borderValues.y
  });
  selectionBorder.attrs.width = borderValues.width;
  selectionBorder.attrs.height = borderValues.height;
  selectionBorder.strokeScaleEnabled(false);

  //add shapes to the group
  Component.add(led, selectionBorder);

  //mouse down event
  Component.on('mousedown', function() {
    ComponentMouseDown(this);
  });

  return Component;
} // end CreateLED()

//This function creates a new pushbutton.
function CreatePushbutton(xpos, ypos, id_nbr) {
  //create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'pb', '', 2, true);

  let swBaseRing = new Konva.Circle({
    radius: 15,
    fill: "rgba(211,211,211,1)",
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1
  });

  let swButton = new Konva.Circle({
    name: "indicator",
    radius: 9,
    fill: defaultFillColor,
    defaultColor: defaultFillColor,
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 2
  });

  //create a selection border for this component and initialize its values
  let selectionBorder = ComponentSelectionBorder();
  selectionBorder.attrs.x = -swBaseRing.attrs.radius;
  selectionBorder.attrs.y = -swBaseRing.attrs.radius;
  selectionBorder.attrs.width = swBaseRing.attrs.radius * 2;
  selectionBorder.attrs.height = swBaseRing.attrs.radius * 2;
  selectionBorder.strokeScaleEnabled(false);

  Component.add(swBaseRing, swButton, selectionBorder);

  Component.addEventListener('mousedown', function() {rndBtnPressedHandler(this);});
  Component.addEventListener('touchstart', function(e) {rndBtnPressedHandler(this); e.stopPropagation();});

  let rndBtnPressedHandler = function(cmp) {

    ComponentMouseDown(cmp);
  };

  return Component;
} // end CreatePushbutton()

//This function creates a new square pushbutton.
function CreateSquarePushbutton(xpos, ypos, id_nbr) {
  //create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'pb_square', '', 2, true);

  let swBase = new Konva.Rect({
    width: 65,
    height: 30,
    fill: defaultFillColor,
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1
  });

  let swButton = new Konva.Rect({
    name: "indicator",
    offsetX: -5,
    offsetY: -5,
    width: 55,
    height: 20,
    fill: "rgba(0,0,0,1)",
    defaultColor: "rgba(0,0,0,1)",
    stroke: "rgba(211,211,211,1)",
    strokeWidth: 2
  });

  let swText = new Konva.Text({
    name: "label",
    text: '',
    fontSize: 10,
    fontFamily: 'Arial',
    fill: "rgba(255,255,255,1)",
    alignText: function()
    {
      swText.offsetX(-swBase.width()/2 + swText.width()/2);
      swText.offsetY(-swBase.height()/2 + swText.height()/2);
    }
  });

  //create a selection border for this component and initialize its values
  let selectionBorder = ComponentSelectionBorder();
  selectionBorder.attrs.width = swBase.getWidth();
  selectionBorder.attrs.height = swBase.getHeight();
  selectionBorder.strokeScaleEnabled(false);

  Component.add(swBase, swButton, swText, selectionBorder);

  //Add button pressed event listeners.
  //If it is a touch event, we don't want the mouse event to be processed also so stop propagation/bubbling.
  //(touch events occur first)
  Component.addEventListener('mousedown', function() {btnPressedHandler(this);});
  Component.addEventListener('touchstart', function(e) {btnPressedHandler(this); e.stopPropagation();});

  let btnPressedHandler = function (cmp) {
    ComponentMouseDown(cmp);
  };

  return Component;
} // end CreateSquarePushbutton()

//This function creates a new toggle switch.
function CreateToggleSwitch(xpos, ypos, id_nbr) {
  //create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'togSw', '', 3, false);

  let swBaseRing = new Konva.Circle({
    radius: 17,
    fill: "rgba(211,211,211,1)",
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1
  });

  let swBaseCntr = new Konva.Circle({
    radius: 7,
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1
  });

  let switchUp = new Konva.Shape({
    name: "up",
    fill: "rgba(211,211,211,1)",
    stroke: "rgba(0,0,0,1)",
    visible: false,
    strokeWidth: 1,
    // a Konva.Canvas renderer is passed into the sceneFunc function
    sceneFunc: function(context, shape)
    {
      context.beginPath();
      context.moveTo(-5, 0);
      context.lineTo(-12, -25);
      context.lineTo(12, -25);
      context.lineTo(5, 0);
      context.bezierCurveTo(5,7,-5,7,-5,0);
      context.moveTo(-12, -25);
      context.lineTo(-12, -30);
      context.lineTo(12,-30);
      context.lineTo(12,-25);
      context.fillStrokeShape(shape);
    }
  });

  let switchDown = switchUp.clone();
  switchDown.name('down');
  switchDown.rotate(180);

  let switchCenter = new Konva.Shape({
    name: "center",
    fill: "rgba(211,211,211,1)",
    stroke: "rgba(0,0,0,1)",
    visible: false,
    strokeWidth: 1,
    // a Konva.Canvas renderer is passed into the sceneFunc function
    sceneFunc: function(context, shape)
    {
      context.beginPath();
      context.moveTo(-12, 5);
      context.lineTo(-12, -5);
      context.lineTo(12, -5);
      context.lineTo(12, 5);
      context.lineTo(-12, 5);
      context.fillStrokeShape(shape);
    }
  });

  //create a selection border for this component and initialize its values
  let selectionBorder = ComponentSelectionBorder();
  selectionBorder.attrs.x = -swBaseRing.attrs.radius;
  selectionBorder.attrs.y = -swBaseRing.attrs.radius;
  selectionBorder.attrs.width = swBaseRing.attrs.radius * 2;
  selectionBorder.attrs.height = swBaseRing.attrs.radius * 2;
  selectionBorder.strokeScaleEnabled(false);

  // Add the component pieces to the group.
  Component.add(swBaseRing, swBaseCntr, switchUp, switchDown, switchCenter, selectionBorder);

  // When adding component from saved file...
  // If this is a momentary toggle switch, change the type
  if (dropMode.togSwMom) {
    Component.attrs.type = "togSwMom";
    switchCenter.visible(true);
  } else {
    switchDown.visible(true);
  }

  Component.addEventListener('mousedown', function() {togglePressedHandler(this);});
  Component.addEventListener('touchstart', function(e) {togglePressedHandler(this); e.stopPropagation();});

  let togglePressedHandler = function(cmp) {

    ComponentMouseDown(cmp);
  };

  return Component;
} // end CreateToggleSwitch()

//This function creates a new timer component.
function CreateTimer(xpos, ypos, id_nbr) {
  //create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'timer', '', 0, true);

  //create a new text object
  let timer = new Konva.Text({
    name: 'indicator',
    text: "0s",
    timerDuration: 0,
    startTime: 0,
    timerRunning: false,
    timerComplete: false,
    fontSize: 20,
    fontFamily: 'Arial',
    fill: "rgba(255,255,255,1)",
    defaultColor: "rgba(255,255,255,1)",

    getBorderValues: function()
    {
      let x, y, width, height;

      x = 0;
      y = 0;
      width = timer.getWidth();
      height = timer.getHeight();

      return {
        x: x,
        y: y,
        width: width,
        height: height
      };
    }
  });

  //create a selection border for this component and initialize its values
  let selectionBorder = ComponentSelectionBorder();
  selectionBorder.attrs.width = timer.getWidth();
  selectionBorder.attrs.height = timer.getHeight();
  selectionBorder.strokeScaleEnabled(false);

  // Add the component pieces to the group.
  Component.add(timer, selectionBorder);

  //mouse down event
  Component.on('mousedown', function() {
    ComponentMouseDown(this);
  });

  return Component;
} // end CreateTimer()

//This function creates a new text component.
function CreateText(xpos, ypos, id_nbr) {
  //create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'text', '', 0, true);

  //create a new text object
  let cmpText = new Konva.Text({
    name: 'indicator',
    text: 'text',
    fontSize: 14,
    fontFamily: 'Arial',
    fill: "rgba(255,0,0,1)",
    defaultColor: "rgba(255,0,0,1)",
    align: lastTextAlign,

    alignText: function()
    {
      switch (cmpText.align()) {
        case "left":
            cmpText.getParent().offsetX(0);
            break;
        case "center":
            cmpText.getParent().offsetX(cmpText.getWidth()/2);
        break;
        case "right":
            cmpText.getParent().offsetX(cmpText.getWidth());
        break;
      }
    },

    getBorderValues: function()
    {
      let x, y, width, height;

      x = 0;
      y = 0;
      width = cmpText.getWidth();
      height = cmpText.getHeight();

      return {
        x: x,
        y: y,
        width: width,
        height: height
      };
    }
  });

  //create a selection border for this component and initialize its values
  let selectionBorder = ComponentSelectionBorder();
  selectionBorder.attrs.width = cmpText.getWidth();
  selectionBorder.attrs.height = cmpText.getHeight();
  selectionBorder.strokeScaleEnabled(false);

  // Add the component pieces to the group.
  Component.add(cmpText, selectionBorder);

  //mouse down event
  Component.on('mousedown', function() {
    ComponentMouseDown(this);
  });

  return Component;
} // end CreateText()

//This function creates a new keyswitch.
function CreateKeySwitch(xpos, ypos, id_nbr) {
  //create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'keyswitch', '', 4, false);

  let swBaseRing = new Konva.Circle({
    radius: 25,
    fill: "rgba(211,211,211,1)",
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1
  });

  let swBaseInnerRing = new Konva.Circle({
    radius: 20,
    fill: "rgba(211,211,211,1)",
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1
  });

  //up and initial position
  let position1 = new Konva.Rect({
    name: "pos1",
    fill: "rgba(0,0,0,1)",
    visible: true,
    strokeWidth: 0,
    x: -4,
    y: -20,
    width: 8,
    height: 25
  });

  //right position
  let position2 = new Konva.Rect({
    name: "pos2",
    fill: "rgba(0,0,0,1)",
    visible: false,
    strokeWidth: 0,
    x: -5,
    y: -4,
    width: 25,
    height: 8
  });

  //left position
  let position3 = new Konva.Rect({
    name: "pos3",
    fill: "rgba(0,0,0,1)",
    visible: false,
    strokeWidth: 0,
    x: -20,
    y: -4,
    width: 25,
    height: 8,
  });

  //create a selection border for this component and initialize its values
  let selectionBorder = ComponentSelectionBorder();
  selectionBorder.attrs.x = -swBaseRing.attrs.radius;
  selectionBorder.attrs.y = -swBaseRing.attrs.radius;
  selectionBorder.attrs.width = swBaseRing.attrs.radius * 2;
  selectionBorder.attrs.height = swBaseRing.attrs.radius * 2;
  selectionBorder.strokeScaleEnabled(false);

  // Add the component pieces to the group.
  Component.add(swBaseRing, swBaseInnerRing, position1, position2, position3, selectionBorder);

  Component.addEventListener('mousedown', function() {ksPressedEventHandler(this);});
  Component.addEventListener('touchstart', function(e) {ksPressedEventHandler(this); e.stopPropagation();});

  let ksPressedEventHandler = function(cmp) {

    ComponentMouseDown(cmp);

  };

  return Component;
} // end CreateKeySwitch()

//This function creates a new rotary switch.
function CreateRotarySwitch(xpos, ypos, id_nbr) {
  // create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'rotaryswitch', '', 7, false);

  let dial = new Konva.Shape({
    name: "dial",
    x: 0,
    y: -2.5,
    fill: "rgba(211,211,211,1)",
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1,

    // a Konva.Context renderer is passed into the sceneFunc function
    sceneFunc: function(context, shape) {
      context.beginPath();
      context.moveTo(-10,-8);
      context.lineTo(0,-28);
      context.lineTo(10,-8);
      context.closePath();
      context.fillStrokeShape(shape);
    }
  });

  let centerline = new Konva.Line({
    points: [0, -30, 0, 0],
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1,
  });

  let rswBaseRing = new Konva.Circle({
    radius: 20,
    fill: "rgba(169,169,169,1)",
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1,
  });

  let rswBaseInnerRing = new Konva.Circle({
    radius: 15,
    fill: "rgba(211,211,211,1)",
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1,
    // strokeEnabled: false,
  });

  //up and initial position
  let dialbase = new Konva.Rect({
    name: "dialbase",
    fill: "rgba(211,211,211,1)",
    visible: true,
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1,
    x: -10,
    y: 10,
    width: 20,
    height: 18
  });

  Component.detent = 0;
  //create a selection border for this component and initialize its values
  let selectionBorder = ComponentSelectionBorder();
  selectionBorder.attrs.x = -rswBaseRing.attrs.radius - 2;
  selectionBorder.attrs.y = -rswBaseRing.attrs.radius - 12;
  selectionBorder.attrs.width = (rswBaseRing.attrs.radius * 2)+4;
  selectionBorder.attrs.height = (rswBaseRing.attrs.radius * 2) + 22;
  selectionBorder.strokeScaleEnabled(false);

  // Add the component pieces to the group.
  Component.add(rswBaseRing, dial, dialbase, rswBaseInnerRing, centerline, selectionBorder);

  Component.addEventListener('mousedown', function() {rsPressedEventHandler(this);});
  Component.addEventListener('touchstart', function(e) {rsPressedEventHandler(this); e.stopPropagation();});

  let rsPressedEventHandler = function(cmp) {

    ComponentMouseDown(cmp);

  };

  return Component;
} // end CreateRotarySwitch()

//***********************************************************************************************************************************************************************
// Components are LEDs, Pushbuttons, toggle switches, text, etc. that can be placed anywhere on the page
// to create a virtual panel that simulates the traditional hardware-based CLCPs.
function AddComponent(xpos, ypos, id_nbr) {
  let Component;
  let configureComponent;

  if (dropMode.quadrilateral) {
    Component = CreateQuadrilateral(xpos, ypos, id_nbr);
    configureComponent = true;
  }

  //create an arrow shape
  if (dropMode.arrow) {
    Component = CreateArrow(xpos, ypos, id_nbr);
    configureComponent = true;
  }

  //create an LED
  if (dropMode.LED) {
    Component = CreateLED(xpos, ypos, id_nbr);
    configureComponent = true;
  }

  //creat a round pushbutton
  if (dropMode.pb) {
    Component = CreatePushbutton(xpos, ypos, id_nbr);
    configureComponent = true;
  }

  //create a square pushbutton
  if (dropMode.pb_square) {
    Component = CreateSquarePushbutton(xpos, ypos, id_nbr);
    configureComponent = true;
  }

  //create a toggle switch
  if (dropMode.togSw || dropMode.togSwMom) {
    Component = CreateToggleSwitch(xpos, ypos, id_nbr);
    configureComponent = true;
  }

  //create a keyswitch
  if (dropMode.keyswitch) {
    Component = CreateKeySwitch(xpos, ypos, id_nbr);
    configureComponent = true;
  }

  // create a rotaryswitch
  if (dropMode.rotaryswitch) {
    Component = CreateRotarySwitch(xpos, ypos, id_nbr);
    configureComponent = true;
  }

  //create text
  if (dropMode.text) {
    Component = CreateText(xpos, ypos, id_nbr);
    configureComponent = true;
    Component.getChildren()[0].attrs.alignText();
  }

  //create timer text
  if (dropMode.timer) {
    Component = CreateTimer(xpos, ypos, id_nbr);
    configureComponent = true;
  }

  //If a new component was added, configure its properties and events
  if (configureComponent) {

    InitializeComponentIndications(Component);
    InitializeComponentControls(Component);

    //add the component to the layer
    component_layer.add(Component);
    //component_layer.draw();
  }

  //return component object
  return Component;
} // end AddComponent()

function InitializeComponentIndications(comp) {
  // Initialize the indication properties for this component
  // Assign default color to each indication based on original color palette).
  for (let i=0; i < stdColors.length * 2; i++) {
    comp.attrs["indication" + i] = "0,0," + colorVals[i % stdColors.length];
  }
} // InitializeComponentIndications()

function InitializeComponentControls(comp) {
  // Initialize the control properties for this component
  for (let i=0; i < comp.attrs.nbrCtrls; i++) {
    comp.attrs["control" + i] = "0";
  }
} // end InitializeComponentControls()

function CheckLockStatus(component) {
  //Check if passcode is defined and/or assigned for this control component
  //and then determine if it is locked.
  if (typeof component.attrs.passcode === 'undefined' ||
      component.attrs.passcode === null ||
      component.attrs.passcode === '') {
    //variable is undefined, null, or unassigned so unlock the control component
    component.attrs.unlocked = true;
  } else if (component.attrs.unlocked === false && initLatchedSwitches === false) {
    //if a passcode is defined for this component and the component is locked...

  }
} // end CheckLockStatus()

const hasColorInputSupport = () => {
  try {
      const input = document.createElement('input');
      input.type = "color";
      input.value = "!";
      return input.type === "color" && input.value !== "!";
  } catch (e) {
      return false;
  };
};

function ComponentProperties(componentObj) {
  // Build html string for component properties dialog box.
  // This string will set the innerHTML of the properties dialog box.
  if (selectedId !== "") {
    let propertiesStr = "";
    let toggleSw = false;

    // Create common properties options.
    propertiesStr =
    '<p class="prop-p">X Position:' +
    '<input id="x-pos-value" type="number" style="width: 208px; float: right" onchange="UpdateComponentPosition(\'' + componentObj.attrs.id + '\')"/></p>' +
    '<p class="prop-p">Y Position:' +
    '<input id="y-pos-value" type="number" style="width: 208px; float: right" onchange="UpdateComponentPosition(\'' + componentObj.attrs.id + '\')"/></p>' +
    '<p class="prop-p">Layer: (Max/Top:' + (getMaxZIndex()+1) + ')' +
    '<input id="z-value-front" type="button" value="Top" style="width: 70px; float: right" onclick="BringComponentToFront(\'' + componentObj.attrs.id + '\')"/>' +
    '<input id="z-value" type="number" min="1" step="1" style="width: 69px; float: right; text-align: center;" onchange="UpdateComponentZIndex(\'' + componentObj.attrs.id + '\')"/>' +
    '<input id="z-value-back" type="button" value="Bottom" style="width: 70px; float: right" onclick="SendComponentToBack(\'' + componentObj.attrs.id + '\')"/></p>';

    // Create PS_Square dimensions property options
    if (componentObj.attrs.type === "pb_square") {
      propertiesStr += '<p class="prop-p">Height:' +
      '<input id="squarepb-height" type="number" min="20" onchange="UpdateSquarePBDimensions(\'' + componentObj.attrs.id + '\')"/></p>' +
      '<p class="prop-p">Width:' +
      '<input id="squarepb-width" type="number" min="20" onchange="UpdateSquarePBDimensions(\'' + componentObj.attrs.id + '\')"/></p>';
    }

    // Create Quadrilateral point coordinate property options.
    if (componentObj.attrs.type === "quadrilateral") {
      propertiesStr += '<p class="prop-p">Point 1 (x,y):' +
      '<input id="point1y-value" type="number" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/>' +
      '<input id="point1x-value" type="number" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/></p>' +
      '<p class="prop-p">Point 2 (x,y):' +
      '<input id="point2y-value" type="number" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/>' +
      '<input id="point2x-value" type="number" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/></p>' +
      '<p class="prop-p">Point 3 (x,y):' +
      '<input id="point3y-value" type="number" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/>' +
      '<input id="point3x-value" type="number" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/></p>' +
      '<p class="prop-p">Point 4 (x,y):' +
      '<input id="point4y-value" type="number" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/>' +
      '<input id="point4x-value" type="number" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/></p>';
    }

    // Add the rotation property for quadrilateral or arrow component.
    if (componentObj.attrs.type === "quadrilateral" || componentObj.attrs.type === "arrow") {
      propertiesStr += '<p class="prop-p">Rotation:' +
      '<input id="rotation-value" type="number" style="width: 208px; float: right" onchange="RotateComponent(\'' + componentObj.attrs.id + '\')"/></p>';
    }

    if (componentObj.attrs.type !== "text" && componentObj.attrs.type !== "timer") {
      propertiesStr += '<p class="prop-p">Scale:' +
      '<input id="scale-value" type="number" style="width: 208px; float: right" onchange="UpdateScaleValue(\'' + componentObj.attrs.id + '\')"/></p>';
    }

    //if the component is not text, add line properties
    if (componentObj.attrs.type === "LED" || componentObj.attrs.type === "quadrilateral" || componentObj.attrs.type === "arrow") {
      propertiesStr += '<p class="prop-p">Line Width:' +
        '<input id="line-width-value" type="number" style="width: 208px; float: right" onchange="UpdateLineWidth(\'' + componentObj.attrs.id + '\')"/></p>';

      if (hasColorInputSupport()) {

        propertiesStr += '<p class="prop-p">Line Color:' +
        '<label class="transp-chkbx-lbl" for="line-transp-chkbx">Transparent</label>' +
        '<input type="checkbox" id="line-transp-chkbx" onchange="UpdateLineColor(\'' + componentObj.attrs.id + '\')"/>' +
        '<input type="color" class="color-input" id="line-color-input" onchange="UpdateLineColor(\'' + componentObj.attrs.id + '\')"/></p>';

        propertiesStr += '<p class="prop-p">Fill Color:' +
        '<label class="transp-chkbx-lbl" for="fill-transp-chkbx">Transparent</label>' +
        '<input type="checkbox" id="fill-transp-chkbx" onchange="UpdateDefaultFillColor(\'' + componentObj.attrs.id + '\')"/>' +
        '<input type="color" class="color-input" id="fill-color-input" onchange="UpdateDefaultFillColor(\'' + componentObj.attrs.id + '\')"/></p>';

      } else {

        propertiesStr += '<p class="prop-p">Line Color:' +
        '<select id="line-color-input" style="width: 210px; float: right" onchange="UpdateLineColor(\'' + componentObj.attrs.id + '\')">';
        // Create color choices
        for (let i = 0; i < stdColors.length; i++) {
          propertiesStr += '<option value="' + colorVals[i] + '">' + stdColors[i] + '</option>';
        }
        propertiesStr += '</select></p>';

        propertiesStr += '<p class="prop-p">Fill Color:' +
        '<select id="fill-color-input" style="width: 210px; float: right" onchange="UpdateDefaultFillColor(\'' + componentObj.attrs.id + '\')">';
        // Create color choices
        for (let i = 0; i < stdColors.length; i++) {
          propertiesStr += '<option value="' + colorVals[i] + '">' + stdColors[i] + '</option>';
        }
        propertiesStr += '</select></p>';
      }
    }

    if (componentObj.attrs.type === "text" || componentObj.attrs.type === "pb_square")
    {
      propertiesStr += '<p class="prop-p">Text:' +
      '<input id="text-value" type="text" style="width: 208px; float: right" onchange="UpdateTextValue(\'' + componentObj.attrs.id + '\')"/></p>';
    }

    if (componentObj.attrs.type === "text") {
      // Create textbox alignment options.
      propertiesStr += '<p class="prop-p">Text Box Alignment:' +
      '<select id="text-box-align" style="width: 210px; float: right" onchange="ChangeTextBoxAlignment(\'' + componentObj.attrs.id + '\')"></p>';
      propertiesStr += '<option>left</option><option>center</option><option>right</option></select>';
    }

    if (componentObj.attrs.type === "text" || componentObj.attrs.type === "timer" ||  componentObj.attrs.type === "pb_square") {
      propertiesStr += '<p class="prop-p">Text Size:' +
      '<select id="text-size" style="width: 210px; float: right" onchange="ChangeTextSize(\'' + componentObj.attrs.id + '\')">';

      // Create text size options.
      for (let i=10; i <= 44; i=i+2) {
        propertiesStr += '<option>' + i + '</option>';
      }
      propertiesStr += '</select></p>';

      if (componentObj.attrs.type === "text" || componentObj.attrs.type === "timer") {

        propertiesStr += '<p class="prop-p">' +
          'Rotation:<input id="rotation-value" type="number" style="width: 208px; float: right" onchange="RotateComponent(\'' + componentObj.attrs.id + '\')"/></p>';

        if (hasColorInputSupport()) {

          propertiesStr += '<p class="prop-p">Text Color:' +
          '<label class="transp-chkbx-lbl" for="text-transp-chkbx">Transparent</label>' +
          '<input type="checkbox" id="text-transp-chkbx" onchange="UpdateDefaultTextColor(\'' + componentObj.attrs.id + '\')"/>' +
          '<input type="color" class="color-input" id="text-color-input" onchange="UpdateDefaultTextColor(\'' + componentObj.attrs.id + '\')"/></p>';

        } else {

          propertiesStr += '<p class="prop-p">Text Color:' +
          '<select id="text-color-input" onchange="UpdateDefaultTextColor(\'' + componentObj.attrs.id + '\')">';
          // Create color choices
          for (let i = 0; i < stdColors.length; i++) {
            propertiesStr += '<option value="' + colorVals[i] + '">' + stdColors[i] + '</option>';
          }
          propertiesStr += '</select></p>';
        }
      }

      if (componentObj.attrs.type === "pb_square") {

        if (hasColorInputSupport()) {

          propertiesStr += '<p class="prop-p">Text Color:' +
          '<label class="transp-chkbx-lbl" for="text-transp-chkbx">Transparent</label>' +
          '<input type="checkbox" id="text-transp-chkbx" onchange="UpdateTextColor(\'' + componentObj.attrs.id + '\')"/>' +
          '<input type="color" class="color-input" id="text-color-input" onchange="UpdateTextColor(\'' + componentObj.attrs.id + '\')"/></p>';

        } else {

          propertiesStr += '<p class="prop-p">Text Color:' +
          '<select id="text-color-input" style="width: 210px; float: right" onchange="UpdateTextColor(\'' + componentObj.attrs.id + '\')">';
          // Create color choices
          for (let i = 0; i < stdColors.length; i++) {
            propertiesStr += '<option value="' + colorVals[i] + '">' + stdColors[i] + '</option>';
          }
          propertiesStr += '</select></p>';
        }
      }
    }

    if (componentObj.attrs.type === "timer") {
      propertiesStr += '<p class="prop-p">Timer Duration:' +
      '<input id="timer-duration-value" type="number" style="width: 208px; float: right" onchange="UpdateTimerDuration(\'' + componentObj.attrs.id + '\')"/></p>';
    }

    //If this is a toggle switch, show option for setting latching or momentary
    if (componentObj.attrs.type === "togSw" || componentObj.attrs.type === "togSwMom") {
      let momentary = "";
      let rotated = "";

      if (componentObj.attrs.type === "togSwMom") {
        momentary = "checked";
      }

      if (componentObj.attrs.rotation == 90) {
        rotated = "checked";
      }

      propertiesStr += '<p class="prop-p">Momentary (ON-OFF-ON):' +
      '<input type="checkbox"' + momentary + ' style="width: 20px; float: right" onchange="SetToggleSwitchType(\'' + componentObj.attrs.id + '\',this.checked)"/></p>' +
      '<p class="prop-p">Rotate 90 degrees:' +
      '<input type="checkbox"' + rotated + ' style="width: 20px; float: right" onchange="RotateToggleSwitch(\'' + componentObj.attrs.id + '\',this.checked)"/></p>';

      toggleSw = true;
    }

    // Loop through the number of controls for this component to create its control properties.
    for (let i=0; i < componentObj.attrs.nbrCtrls ; i++) {
      let labelNbr = i + 1;
      let descStr = "";

      if (toggleSw) {
        if (labelNbr == 1) {
          descStr = " (Up/Right Position)";
        }
        if (labelNbr == 2) {
          descStr = " (Down/Left Position)";
        }
      }

      if (componentObj.attrs.type === "keyswitch") {
        if (labelNbr == 1) {
          descStr = " (Up Position)";
        }
        if (labelNbr == 2) {
          descStr = " (Right Position)";
        }
        if (labelNbr == 3) {
          descStr = " (Left Position)";
        }
      }

      if (componentObj.attrs.type === "rotaryswitch") {
        switch(labelNbr) {
          case 1:
            descStr = " (12 o'clock)";
          break;
          case 2:
            descStr = " (2 o'clock)";
          break;
          case 3:
            descStr = " (4 o'clock)";
          break;
          case 4:
            descStr = " (6 o'clock)";
          break;
          case 5:
            descStr = " (8 o'clock)";
          break;
          case 6:
            descStr = " (10 o'clock)";
          break;
        }
      }

      //the last control is used for the unlock status
      if (labelNbr == componentObj.attrs.nbrCtrls) {
        descStr = " (Unlock Status)";
      }

      propertiesStr += '<p class="prop-p">Control Status ' + labelNbr + descStr + ': ' +
      '<select id="control' + i + '-lbl-opts" style="float: right; width: 210px" onchange="SetComponentControlProperty(\'' + componentObj.attrs.id + '\',\'control' + i + '\')">' +
      controlOptionsStr +
      '</select></p>';
    }

    if (componentObj.attrs.hasIndications) {
      //Add flash rate property
      propertiesStr += '<p class="prop-p">Flash Time (ms):' +
      '<input id="flash-time-value" type="number" style="width: 208px; float: right" onchange="UpdateComponentFlashTime(\'' + componentObj.attrs.id + '\')"/></p>';

      // -for now, loop through the indications.
      // -soon Add the assigned indications (and selectors to add/remove indication assignments).
      if (hasColorInputSupport()) {

        //Add the labels and drop down lists to select status names for the colored indication. Also add the lists to select the state option.
        //The state option determines whether the color is displayed when the bit state is true or when it is false.
        for (let indNbr=0; indNbr < stdColors.length * 2; indNbr++) {

          if (indNbr < stdColors.length) {
            propertiesStr += '<p class="prop-p">Indication:';
          } else {
            propertiesStr += '<p class="prop-p"> Flash Ind:';
          }
          propertiesStr += '<select id="indication'+ indNbr + '-lbl-opts" onchange="SetComponentIndicationProperty(\'' + componentObj.attrs.id + '\',\'indication' + indNbr + '\')">' +
          indicationOptionsStr +
          '</select>' +
          '<select id="indication'+ indNbr + '-val-opts" onchange="SetComponentIndicationProperty(\'' + componentObj.attrs.id + '\',\'indication' + indNbr + '\')">' +
          '<option>True</option><option>False</option>' +
          '</select>' +
          '<label class="transp-chkbx-lbl" for="indication' + indNbr + '-transp-chkbx">Transparent</label>' +
          '<input type="checkbox" id="indication' + indNbr + '-transp-chkbx" onchange="SetComponentIndicationProperty(\'' + componentObj.attrs.id + '\',\'indication' + indNbr + '\')"/>' +
          '<input type="color" id="indication' + indNbr + '-color-input" onchange="SetComponentIndicationProperty(\'' + componentObj.attrs.id + '\',\'indication' + indNbr + '\')"/></p>';
        }
      } else { // Color picker input not supported in browser, provide color dropdowns instead.
        //Add the labels and drop down lists to select status names for the colored indication. Also add the lists to select the state option.
        //The state option determines whether the color is displayed when the bit state is true or when it is false.
        let indNbr = 0;
        for (let i=0; i < stdColors.length * 2; i++) {

          if (i < stdColors.length) {
            propertiesStr += '<p class="prop-p">Indication:';
          } else {
          propertiesStr += '<p class="prop-p">Flash Ind:';
          }

          propertiesStr += '<select id="indication'+ indNbr + '-lbl-opts" onchange="SetComponentIndicationProperty(\'' + componentObj.attrs.id + '\',\'indication' + indNbr + '\')">' +
          indicationOptionsStr +
          '</select>' +
          '<select id="indication'+ indNbr + '-val-opts" onchange="SetComponentIndicationProperty(\'' + componentObj.attrs.id + '\',\'indication' + indNbr + '\')">' +
          '<option>True</option><option>False</option>' +
          '</select>' +
          '<select id="indication' + indNbr + '-color-input" style="width:110px; float: none;" "SetComponentIndicationProperty(\'' + componentObj.attrs.id + '\',\'indication' + indNbr + '\')">';
          // Create color choices
          for (let i = 0; i < stdColors.length; i++) {
            propertiesStr += '<option value="' + colorVals[i] + '">' + stdColors[i] + '</option>';
          }
          propertiesStr += '</select></p>';

          indNbr++;
        }
      }
    }

    // Build the final string and assign it to the properties dialog box.
    document.getElementById("properties_dialog").innerHTML = propertiesStr;

    // Now that the inner html for the properties is created,
    // get the property values for the current component.
    document.getElementById("x-pos-value").value = Math.round(componentObj.attrs.x);
    document.getElementById("y-pos-value").value = Math.round(componentObj.attrs.y);
    document.getElementById('z-value').value = componentObj.zIndex() + 1;

    let colorVal = "";

    if (componentObj.attrs.type === "text" || componentObj.attrs.type === "pb_square") {
      document.getElementById("text-value").value = componentObj.find('Text')[0].attrs.text;
    }

    if (componentObj.attrs.type === "text") {
      document.getElementById("text-box-align").value = componentObj.find('Text')[0].attrs.align;
    }

    if (componentObj.attrs.type === "text" || componentObj.attrs.type === "timer") {
      let cmpAttr = componentObj.find('Text')[0].attrs;

      document.getElementById("text-size").value = cmpAttr.fontSize;

      if (hasColorInputSupport()) {
        colorVal = RGBAToHexA(cmpAttr.defaultColor);

        if (colorVal.substring(7) === "00") {
          document.getElementById("text-transp-chkbx").checked = true;
          document.getElementById("text-color-input").style.visibility = "hidden";
        }
        colorVal = colorVal.substring(0, 7);
      } else {
        colorVal = cmpAttr.defaultColor;
      }

      document.getElementById("text-color-input").value = colorVal;

    } else {
      document.getElementById("scale-value").value = Math.round(componentObj.attrs.scaleX * 100);
    }

    if (componentObj.attrs.type === "pb_square") {
      let cmpAttr = componentObj.find('Text')[0].attrs;
      document.getElementById("text-size").value = cmpAttr.fontSize;

      if (hasColorInputSupport()) {
        colorVal = RGBAToHexA(cmpAttr.fill);

        if (colorVal.substring(7) === "00") {
          document.getElementById("text-transp-chkbx").checked = true;
          document.getElementById("text-color-input").style.visibility = "hidden";
        }
        colorVal = colorVal.substring(0, 7);
      } else {
        colorVal = cmpAttr.fill;
      }
      document.getElementById("text-color-input").value = colorVal;

      // Now get the current dimensions.
      cmpAttr = componentObj.find('Rect')[0];
      document.getElementById("squarepb-height").value = cmpAttr.height();
      document.getElementById("squarepb-width").value = cmpAttr.width();

    }

    if (componentObj.attrs.type === "quadrilateral") {

      let quad = componentObj.find('.indicator')[0].attrs;
      document.getElementById("point1x-value").value = quad.point1x;
      document.getElementById("point1y-value").value = quad.point1y;
      document.getElementById("point2x-value").value = quad.point2x;
      document.getElementById("point2y-value").value = quad.point2y;
      document.getElementById("point3x-value").value = quad.point3x;
      document.getElementById("point3y-value").value = quad.point3y;
      document.getElementById("point4x-value").value = quad.point4x;
      document.getElementById("point4y-value").value = quad.point4y;
    }

    if (componentObj.attrs.type === "timer") {
      document.getElementById("timer-duration-value").value = componentObj.find('Text')[0].attrs.timerDuration;
    }

    if (componentObj.attrs.type === "text" || componentObj.attrs.type === "quadrilateral" ||
      componentObj.attrs.type === "arrow" || componentObj.attrs.type === "timer") {

      document.getElementById("rotation-value").value = componentObj.attrs.rotation;
    }

    if (componentObj.attrs.type === "LED" || componentObj.attrs.type === "quadrilateral" || componentObj.attrs.type === "arrow") {
      let cmpAttr = componentObj.find('.indicator')[0].attrs;

      document.getElementById("line-width-value").value = cmpAttr.strokeWidth;

      if (hasColorInputSupport()) {
        colorVal = RGBAToHexA(cmpAttr.stroke);

        if (colorVal.substring(7) === "00") {
          document.getElementById("line-transp-chkbx").checked = true;
          document.getElementById("line-color-input").style.visibility = "hidden";
        }
        colorVal = colorVal.substring(0,7);
      }  else {
        colorVal = cmpAttr.stroke;
      }

      document.getElementById("line-color-input").value = colorVal;

      if (hasColorInputSupport()) {
        colorVal = RGBAToHexA(cmpAttr.defaultColor);

        if (colorVal.substring(7) === "00") {
          document.getElementById("fill-transp-chkbx").checked = true;
          document.getElementById("fill-color-input").style.visibility = "hidden";
        }
        colorVal = colorVal.substring(0,7);
      } else {
        colorVal = cmpAttr.defaultColor;
      }

      document.getElementById("fill-color-input").value = colorVal;
    }

    for (let i=0; i < componentObj.attrs.nbrCtrls; i++) {
      document.getElementById("control" + i + "-lbl-opts").selectedIndex = componentObj.attrs["control" + i];
    }

    if (componentObj.attrs.hasIndications) {
      document.getElementById("flash-time-value").value = componentObj.attrs.flashTime;

      for (let indIdx = 0; indIdx < stdColors.length * 2; indIdx++) {

        let indObjAttrs = componentObj.attrs["indication" + indIdx].split(",");
        let state = document.getElementById("indication" + indIdx + "-val-opts");
        let color = document.getElementById("indication" + indIdx + "-color-input");
        let transp = document.getElementById("indication" + indIdx + "-transp-chkbx");

        let labelIdx = indObjAttrs.shift();

        document.getElementById("indication" + indIdx + "-lbl-opts").selectedIndex = labelIdx;
        state.selectedIndex = indObjAttrs.shift();

        let colorVal = indObjAttrs.join();
        let vis = (labelIdx == 0) ? "hidden" : "visible";

        state.style.visibility = vis;
        color.style.visibility = vis;

        if (transp == null) {
          color.value = colorVal;
        } else {

          transp.style.visibility = vis;
          transp.labels[0].style.visibility = vis;

          colorVal = RGBAToHexA(colorVal);

          if (colorVal.substring(7) === "00") {
            transp.checked = true;
            color.style.visibility = "hidden";
          }
          colorVal = colorVal.substring(0,7);
        }

        color.value = colorVal;
      }
    }

  } // end if (selectedId !== "")
} // end ComponentProperties()

function SetToggleSwitchType(id, checked) {
  let componentObj = stage.find('#'+id)[0];

  if (checked) {
    componentObj.attrs.type = "togSwMom";
    componentObj.find('.up')[0].visible(false);
    componentObj.find('.down')[0].visible(false);
    componentObj.find('.center')[0].visible(true);
  } else {
    componentObj.attrs.type = "togSw";
    componentObj.find('.up')[0].visible(false);
    componentObj.find('.down')[0].visible(true);
    componentObj.find('.center')[0].visible(false);
  }

  component_layer.draw();
} // end SetToggleSwitchType()

function SetComponentControlProperty(id, property) {
  //This function is called when control status name (bit assignement) is changed to update the saved values.
  let statusListId = property + "-lbl-opts";
  let componentObj = stage.find('#'+id)[0];
  componentObj.attrs[property] = document.getElementById(statusListId).selectedIndex;
} // end SetComponentControlProperty()

function SetComponentIndicationProperty(id, property) {
  // This function is called when an indication color, status name, or state option is changed to update the saved values.
  // The state option determines whether the color is displayed when the bit state is true or when it is false.
  let textObj = stage.find('#'+id)[0];
  let state = document.getElementById(property + "-val-opts");
  let color = document.getElementById(property + "-color-input");
  let transp = document.getElementById(property + "-transp-chkbx");

  let labelIdx = document.getElementById(property + "-lbl-opts").selectedIndex;
  let stateIdx = state.selectedIndex;

  // Update the visibility based on having an assigned label or not.
  let vis = (labelIdx == 0) ? "hidden" : "visible";

  state.style.visibility = vis;
  color.style.visibility = vis;

  if (transp == null) {
    color = color.value;
  } else {

    transp.style.visibility = vis;
    transp.labels[0].style.visibility = vis;

    if (transp.checked) {
      color.style.visibility = "hidden";
      transp = '00';
    } else {
      transp = 'FF';
    }
    color = hexAToRGBA(color.value + transp);
  }

  textObj.attrs[property] = labelIdx + "," + stateIdx + "," + color;
} // end SetComponentIndicationProperty()

function UpdateTextValue(id) {
  let textObj = stage.find('#'+id)[0].find('Text')[0];
  textObj.setText(document.getElementById("text-value").value);

  if (textObj.getParent().attrs.type === 'pb_square' || textObj.getParent().attrs.type === 'text') {
    textObj.attrs.alignText();
  }

  //update the selection border to account for line width
  stage.find('#'+id)[0].find('.selection')[0].attrs.update();

  component_layer.draw();
} // end UpdateTextValue()

function UpdateDefaultTextColor(id) {
  const component = stage.find('#'+id)[0].find('Text')[0];
  let color = document.getElementById("text-color-input");
  let transp = document.getElementById("text-transp-chkbx");

  if (transp != null ) {
    if (transp.checked) {
      color.style.visibility = "hidden";
      transp = '00';
    } else {
      transp = 'FF';
      color.style.visibility = "visible";
    }
    color = hexAToRGBA(color.value + transp);
  } else {
    color = color.value;
  }

  component.attrs.defaultColor = color;
  component_layer.draw();
} // end UpdateDefaultTextColor()

function UpdateTextColor(id) {
  const component = stage.find('#'+id)[0].find('Text')[0];
  let color = document.getElementById("text-color-input");
  let transp = document.getElementById("text-transp-chkbx");

  if (transp != null ) {
    if (transp.checked) {
      color.style.visibility = "hidden";
      transp = "00";
    } else {
      transp = "FF";
      color.style.visibility = "visible";
    }
    color= hexAToRGBA(color.value + transp);
  } else {
    color = color.value;
  }

  component.attrs.fill = color;
  component_layer.draw();
} // end UpdateTextColor()

function UpdateLineColor(id) {
  const component = stage.find('#'+id)[0].find('.indicator')[0];
  let color = document.getElementById("line-color-input");
  let transp = document.getElementById("line-transp-chkbx");

  if (transp != null ) {
    if (transp.checked) {
      color.style.visibility = "hidden";
      transp = "00";
    } else {
      transp = "FF";
      color.style.visibility = "visible";
    }
    color = hexAToRGBA(color.value + transp);
  } else {
    color = color.value;
  }

  component.stroke(color);
  component_layer.draw();
} // end UpdateLineColor()

function UpdateDefaultFillColor(id) {
  const component = stage.find('#'+id)[0].find('.indicator')[0];
  let color = document.getElementById("fill-color-input");
  let transp = document.getElementById("fill-transp-chkbx")

  if (transp != null ) {
    if (transp.checked) {
      color.style.visibility = "hidden";
      transp = "00";
    } else {
      transp = "FF";
      color.style.visibility = "visible";
    }
    color = hexAToRGBA(color.value + transp);
  } else {
    color = color.value;
  }

  component.attrs.defaultColor = color;
  component_layer.draw();
 }// end UpdateDefaultFillColor()

function ChangeTextBoxAlignment(id) {
  let textObj = stage.find('#'+id)[0].find('Text')[0];
  textObj.align(document.getElementById("text-box-align").value);
  lastTextAlign = textObj.align();
  textObj.attrs.alignText();
} // end ChangeTextBoxAlignment()

function ChangeTextSize(id) {
  let textObj = stage.find('#'+id)[0].find('Text')[0];
  textObj.fontSize(document.getElementById("text-size").value);

  //refresh the text value so that selection border resizing will work
  textObj.setText(textObj.attrs.text);

  if (textObj.getParent().attrs.type === 'pb_square' ||
      textObj.getParent().attrs.type === 'text') {
    textObj.attrs.alignText();
  }

  //update the selection border to account for line width
  stage.find('#'+id)[0].find('.selection')[0].attrs.update();

  component_layer.draw();
} // end ChangeTextSize()

function IncreaseBackgroundImageSize() {
  //if there is an image, resize it based on the current scale factor
  if (bgImage_layer.hasChildren()) {
    let bgImg = bgImage_layer.getChildren()[0];
    let sz = bgImg.size();
    bgImg.size({
      width: sz.width + 10,
      height: sz.height + (10 * sz.height/sz.width)
    });
  }
  bgImage_layer.draw();
} // end IncreaseBackgroundImageSize()

function DecreaseBackgroundImageSize() {
  //if there is an image, resize it based on the current scale factor
  if (bgImage_layer.hasChildren()) {
    let bgImg = bgImage_layer.getChildren()[0];
    let sz = bgImg.size();
    bgImg.size({
      width: sz.width - 10,
      height: sz.height - (10 * sz.height/sz.width)
    });
  }
  bgImage_layer.draw();
} // end DecreaseBackgroundImageSize()

function RemoveBackgroundImage() {
  if (bgImage_layer.hasChildren()) {
    bgImage_layer.removeChildren();
    bgImage_layer.clear();
    bgImage_layer.draw();
  }
} // end RemoveBackgroundImage()

function TogglePositionLock() {

  const allObjects = component_layer.getChildren();

  if (TogglePositionLock.positionLocked == true) {
    TogglePositionLock.positionLocked = false;
    allObjects.draggable(true);
  } else {
    TogglePositionLock.positionLocked = true;
    allObjects.draggable(false);
  }
} // end TogglePositionLock()

function UpdateComponentPosition(id)
{
  let component = stage.find('#'+id)[0];

  component.position({
    x: Math.floor(Number(document.getElementById("x-pos-value").value)),
    y: Math.floor(Number(document.getElementById("y-pos-value").value))
  });
  component_layer.draw();
} // end UpdateComponentPosition()

function getMaxZIndex() {
  //find the maximum zIndex.
  let maxZ = 0;
  for (let j=0; j < component_layer.children.length; j++) {
    let cmpObj = component_layer.children[j];
    if (cmpObj.zIndex() > maxZ) {
      maxZ = cmpObj.zIndex();
    }
  }
  return maxZ;
} // end getMaxZIndex()

function UpdateComponentZIndex(id) {
  let comp = stage.find('#'+id)[0];
  let z = Math.round(document.getElementById("z-value").value)-1;
  let maxZ = getMaxZIndex();

  if (z < 0) {
    z = 0;
  } else if (z > maxZ) {
    z = maxZ;
  }

  comp.zIndex(z);
  document.getElementById("z-value").value = z+1;

  component_layer.draw();
} // end UpdateComponentZIndex()

function BringComponentToFront(id) {
  let component = stage.find('#'+id)[0];
  component.zIndex(getMaxZIndex());
  document.getElementById("z-value").value = component.zIndex()+1;
  component_layer.draw();
} // end BringComponentToFront()

function SendComponentToBack(id) {
  let component = stage.find('#'+id)[0];
  component.zIndex(0);
  document.getElementById("z-value").value = 1;
  component_layer.draw();
} // end SendComponentToBack()

function UpdateLineWidth(id) {
  if (Number.isInteger(Number(document.getElementById("line-width-value").value))) {
    let component = stage.find('#'+id)[0].find('.indicator')[0];

    if (document.getElementById("line-width-value").value <= 0) {
      document.getElementById("line-width-value").value = 0;
      component.strokeEnabled(false);
    } else {
      component.strokeEnabled(true);
    }

    if (document.getElementById("line-width-value").value > 20) {
      document.getElementById("line-width-value").value = 20;
    }

    component.strokeWidth(document.getElementById("line-width-value").value);

    //update the selection border to account for line width
    stage.find('#'+id)[0].find('.selection')[0].attrs.update();

    component_layer.draw();
  }
} // end UpdateLneWidth()

function UpdateScaleValue(id) {
  let component = stage.find('#'+id)[0];

  component.scale({
    x: document.getElementById("scale-value").value/100,
    y: document.getElementById("scale-value").value/100
  });

  //update the selection border to account for line width
  stage.find('#'+id)[0].find('.selection')[0].attrs.update();

  component_layer.draw();
} // end UpdateScaleIndex()

function UpdateSquarePBDimensions(id) {

  let pbComps = stage.find('#'+id)[0].find('Rect');
  let hght = Math.round(document.getElementById("squarepb-height").value);
  let wdth = Math.round(document.getElementById("squarepb-width").value);

  if (hght < 20) {
    hght = 20;
  }

  if (wdth < 20) {
    wdth = 20;
  }

  // Update the input field.
  document.getElementById("squarepb-height").value = hght;
  document.getElementById("squarepb-width").value = wdth;

  // Update the switch base dimensions.
  pbComps[0].height(hght);
  pbComps[0].width(wdth);

  // Update the indicator dimensions.
  pbComps[1].width(pbComps[0].width()-10);
  pbComps[1].height(pbComps[0].height()-10);

  // Update the selection border.
  pbComps[2].width(pbComps[0].width());
  pbComps[2].height(pbComps[0].height());

  // Re-align the text.
  UpdateTextValue(id);
} // end UpdateSquarePBDimensions()

function UpdateQuadrilateralCoordinates(id) {
  let minPixelDiff = 2;

  //get the quad shape from the stage
  let quad = stage.find('#'+id)[0].find('.indicator')[0].attrs;

  //check coordinates...
  //first, round any decimal values
  document.getElementById("point1x-value").value = Math.round(document.getElementById("point1x-value").value);
  document.getElementById("point2x-value").value = Math.round(document.getElementById("point2x-value").value);
  document.getElementById("point3x-value").value = Math.round(document.getElementById("point3x-value").value);
  document.getElementById("point4x-value").value = Math.round(document.getElementById("point4x-value").value);
  document.getElementById("point1y-value").value = Math.round(document.getElementById("point1y-value").value);
  document.getElementById("point2y-value").value = Math.round(document.getElementById("point2y-value").value);
  document.getElementById("point3y-value").value = Math.round(document.getElementById("point3y-value").value);
  document.getElementById("point4y-value").value = Math.round(document.getElementById("point4y-value").value);

  //2x cannot be less than minPixelDiff greater than 1x
  if (Number(document.getElementById("point2x-value").value) - Number(document.getElementById("point1x-value").value) < minPixelDiff) {
    document.getElementById("point2x-value").value = Number(document.getElementById("point1x-value").value) + minPixelDiff;
  }

  //3x cannot be less than minPixelDiff greater than 4x
  if (Number(document.getElementById("point3x-value").value) - Number(document.getElementById("point4x-value").value) < minPixelDiff) {
    document.getElementById("point3x-value").value = Number(document.getElementById("point4x-value").value) + minPixelDiff;
  }

  //4y cannot be less than minPixelDiff greater than 1y
  if (Number(document.getElementById("point4y-value").value) - Number(document.getElementById("point1y-value").value) < minPixelDiff) {
    document.getElementById("point4y-value").value = Number(document.getElementById("point1y-value").value) + minPixelDiff;
  }

  //3y cannot be less than minPixelDiff greater than 2y
  if (Number(document.getElementById("point3y-value").value) - Number(document.getElementById("point2y-value").value) < minPixelDiff) {
    document.getElementById("point3y-value").value = Number(document.getElementById("point2y-value").value) + minPixelDiff;
  }

  quad.point1x = Number(document.getElementById("point1x-value").value);
  quad.point1y = Number(document.getElementById("point1y-value").value);
  quad.point2x = Number(document.getElementById("point2x-value").value);
  quad.point2y = Number(document.getElementById("point2y-value").value);
  quad.point3x = Number(document.getElementById("point3x-value").value);
  quad.point3y = Number(document.getElementById("point3y-value").value);
  quad.point4x = Number(document.getElementById("point4x-value").value);
  quad.point4y = Number(document.getElementById("point4y-value").value);

  //update the selection border since the size has changed
  stage.find('#'+id)[0].find('.selection')[0].attrs.update();

  component_layer.draw();
} // end UpdateQuadrilateralCoordinates()

function RotateComponent(id) {
  let component = stage.find('#'+id)[0];

  if ((document.getElementById("rotation-value").value < 0)||(document.getElementById("rotation-value").value >= 360)) {
    document.getElementById("rotation-value").value = 0;
  }

  component.rotation(document.getElementById("rotation-value").value);
  component_layer.draw();
} // end RotateComponent()

function RotateToggleSwitch(id, checked) {
  let component = stage.find('#'+id)[0];

  if (checked) {
    component.rotation(90);
  } else {
    component.rotation(0);
  }

  component_layer.draw();
} // end RotateToggleSwitch()

function UpdateComponentFlashTime(id) {
  let component = stage.find('#'+id)[0];

  if (document.getElementById("flash-time-value").value < 0) {
    document.getElementById("flash-time-value").value = 0;
  }

  if (document.getElementById("flash-time-value").value > 65535) {
    document.getElementById("flash-time-value").value = 65535;
  }

  component.attrs.flashTime = document.getElementById("flash-time-value").value;
} // end UpdateComponentFlashTime()

function UpdateTimerDuration(id) {
  let component = stage.find('#'+id)[0].find('.indicator')[0];

  if (document.getElementById("timer-duration-value").value < 0) {
    document.getElementById("timer-duration-value").value = 0;
  }

  if (document.getElementById("timer-duration-value").value > 65535) {
    document.getElementById("timer-duration-value").value = 65535;
  }

  component.attrs.timerDuration = document.getElementById("timer-duration-value").value;
  component.setText(component.attrs.timerDuration.toString() + 's');

  //update the selection border to account for line width
  stage.find('#'+id)[0].find('.selection')[0].attrs.update();
} // end UpdateTimerDuration()

function DuplicateSelectionToPanel() {
  let duplicateComponents = [];
  let dx, dy;

  if (selectedComponents.length) {

    // First, ensure no Component drop buttons are active on the main panel.
    window.opener.DeselectDropComponentButtons();

    // Sort selectedComponents by z-index so that the z-order is maintained upon duplication.
    selectedComponents.sort( (a,b) => (a.zIndex() > b.zIndex() ? 1 : -1 ) );

    for (let comp of selectedComponents) {

      if ( document.getElementById('duplicate-position').checked ) {
        dx = -50;
        dy = 0;
      } else {

        // Place at default copy location while keeping relative spacing in the selection.
        // Calcluate the delta to move the first component to the copy coordinates).
        if (dx === undefined) {
          dx = 200 - comp.attrs.x; // DuplicateComponent() shifts the x-coord an additional 50.
          dy = 250 - comp.attrs.y;
        }

        // Snap to selection mouse once panel window gets focus.
        window.opener.SetStickySelection();
      }

      // Move each component by the same amount.
      comp.attrs.x += dx;
      comp.attrs.y += dy;

      let duplicate = window.opener.DuplicateComponent(comp, !document.getElementById('duplicate-props').checked);

      // Restore the original x,y coordinates.
      comp.attrs.x -= dx;
      comp.attrs.y -= dy;

      // Keep list of new duplicated components
      if (duplicate) {
        duplicateComponents.push(duplicate);
      }
    }

    // Clear previously selected components.
    window.opener.clearComponentSelection();

    // Add the 'add/duplicate components' action to the main stage undo buffer
    window.opener.pushActionToUndoBuf({action:'add', cids: duplicateComponents.map(e=>e.attrs.id)});

    // Make the duplicated components the selected components on the main stage.
    for (const comp of duplicateComponents) {
      window.opener.addComponentToSelection(comp, false);
    }

    // Make the 1st component the active component in the selection list.
    window.opener.updateActiveComponent(duplicateComponents[0]);

    // Update the position with the active selected component.
    window.opener.ShowComponentPosition(duplicateComponents[0]);
  }
} // end DuplicateSelectionToPanel()

//***********************************************************************************************************************************************************************
//after the document page has loaded and is ready...
$(document).ready(function() {

  InitGrid();

  InitDialogBoxes();

  // Ensure that components are created with positions locked.
  TogglePositionLock.positionLocked = true;

  // This will periodically call UpdateComponentLayer to draw components as they are added/updated.
  window.requestAnimationFrame(UpdateComponentLayer);

  // Prevent context menu from popping up when user clicks on right mouse button
  document.addEventListener("contextmenu", function(e){
      e.preventDefault();
  }, false);
});

//This function will delete or move selected component(s) using the keyboard.
//This function depends on requestAnimationFrame to redraw the component layer.
$(document).keydown(function (e) {

    switch(e.key) {
      case "Escape":
        ClearSelectedComponents();
      break;

      case "Control":
        ctrlKeyDown = true;
        break;
    }
});

$(document).keyup(function (e){
  switch (e.key) {
    case "Control":
      ctrlKeyDown = false;
    break;
  }
});
