//qnet-client.js
//'use strict';

//This is a web worker. It runs as a different "thread" from the main page UI thread.
//This web worker handles the requests for indications from the server (real QLCP target or developer tool) and sends controls.
//Received indications are parsed and each bit state is saved in the indicationBits array which is posted for the UI every other request interval.
//Each component in the UI is listening for the postmessage event from this web worker.

let requestInterval = 50;                   //this is the interval used for the periodic functions. It sets the base time in milliseconds for all periodicity
let flashTime = 10;                         //this is the flash interval which is multiplied times the requestInterval for determining total time
let flashTimeCntr = 0;                      //interval counter used to determine on/off flash duration
let reqIndFlag = true;                      //flag to determine whether to request indication or flash bits
let linkTimeout = 40;                       //link timeout = this value times the requestInterval
let linkTimeoutIntervalCntr = 0;            //interval counter to determine when to set link state to failed
let linkHealth = false;                     //link health state
let skipRequest = false;                    //this is for developer mode which does not request flash statuses
let indicationData = "";
let lastIndicationData = "";
let flashData = "";
let lastFlashData = "";

let requestActive = false;                  //don't send another request if this flag is true
let devMode = false;
let urlRoot = "";

let numIndicationBits = 0;                  //number of indication bits
let indicationBytes = [];                   //array of indication bytes after indicationData is split
let indicationMaskBytes = [];               //array of indication mask ('known' state) bytes after indicationData is split
let indFlashBytes = [];                     //array of indication flash bytes
let flashMaskBytes = [];
let flashOffState = false;
let indicationBits = [];
let indicationMaskBits = [];
let ready = false;                          //true when ready to request indications from the server

//Add the event listener to process messages received from the UI thread.
//The message received from the main thread contains the following variables:
//   devMode - true when the web GUI is being used in the developer mode
//   url - IP address of the QLCP-NET entered in the browser
//   sendControl - true when the main thread is requesting an update control to be sent (one bit)
//   bitNbr - the bit number associated with the update control
//   bitState - true/false state of the bit associated with the update control
//   sendAllControls - true when the main thread is requesting a full control to be sent (all bits)
//   cntrlBits - array of the control bit states

this.addEventListener("message", function(e){

  if (e.data.devMode)
  {
    devMode = true;
    urlRoot = "http://localhost:" + e.data.serverPort + "/";
  }
  else
  {
    devMode = false;
    urlRoot = "http://" + e.data.url + ":54321/";
  }

  //Set up structures to handle indications (if not already done)
  if (numIndicationBits != e.data.numIndicationBits)
  {
    numIndicationBits = e.data.numIndicationBits;

    let numIndicationBytes = Math.ceil(numIndicationBits / 8);

    indicationBytes = new Array(numIndicationBytes);
    indicationMaskBytes = new Array(numIndicationBytes);
    indFlashBytes = new Array(numIndicationBytes);
    flashMaskBytes = new Uint8Array(numIndicationBytes);

    indicationBits = new Array(numIndicationBits);
    indicationMaskBits = new Array(numIndicationBits);

    //init flash bytes and flash mask
    for (let i = 0; i < indFlashBytes.length; i++)
    {
      indFlashBytes[i] = 0;
      flashMaskBytes[i] = 0xFF;
    }

    //init 'known/unknown' state mask bytes
    //default to indicating that all indication bits are in a 'known' state
    for (let i = 0; i < indicationMaskBytes.length; i++)
    {
      indicationMaskBytes[i] = "FF";        // 0xFF (all bits 'known'/'certain')
    }
  }

  //Send a control if requested by the UI thread.
  if (e.data.sendControl)
  {
    qnetSendControl(e.data.bitNbr, e.data.bitState);
  }

  if (e.data.sendAllControls)
  {
    qnetSendAllControls(e.data.cntrlBits);
  }

  // console.log(e.data);

  //Now that we've received a message from the main thread defining the
  //developer mode and url, we are ready to request indications.
  ready = true;

});

RequestIndications();

function RequestIndications()
{
  let url = "";

  skipRequest = false;

  //if no indication data is received for link timeout time, show link failure icon in header
  if (linkTimeoutIntervalCntr >= linkTimeout)
  {
    linkHealth = false;

    //clear indication data
    indicationData = "";
    lastIndicationData = "";
    flashData = "";
    lastFlashData = "";
  }
  else
  {
    linkTimeoutIntervalCntr++;
  }

  if (ready)
  {
    //if there is not currently an active request...
    if (!requestActive)
    {
      //Assign url based on developer mode or time to request indication bits vs flash bits.
      //Every other request is for flashbits when connected to a QLCP-NET (not in developer mode).
      if (reqIndFlag)
      {
        reqIndFlag = false; //next time request flash bits
        url = urlRoot + "indications/";
      }
      else
      {
        reqIndFlag = true;  //next time request indication bits
        url = urlRoot + "flashbits/";

        //skip flash request if this is developer mode
        if (devMode) skipRequest = true;
      }

      if (!skipRequest)
      {
        requestActive = true;

        //Request the indication statuses
        let Http = new XMLHttpRequest();
        Http.open("GET", url, true);
        Http.send();
        Http.onreadystatechange= function( ) {
          //if request finished and response is ready...
          if (this.readyState === 4)
          {
            if (this.status === 200) //OK
            {
              linkTimeoutIntervalCntr = 0;
              linkHealth = true;
              if (!reqIndFlag)
              {
                //save the message data
                indicationData = Http.responseText;

                //Split the indication data into data bytes and mask bytes
                let indicationDataTokens = indicationData.split("|");

                //Split the data into its separate bytes
                indicationBytes = indicationDataTokens[0].split(",");

                //Extract mask ('known' state) data if present
                if (indicationDataTokens.length > 1) {
                  indicationMaskBytes = indicationDataTokens[1].split(",");
                }
              }
              else
              {
                //save the message data
                flashData = Http.responseText;

                //Split the data into its separate bytes
                indFlashBytes = Http.responseText.split(",");
              }
            }
            requestActive = false;
          }
        };
      }
    }

    if (flashTimeCntr === flashTime)
    {
      if (flashOffState === false)
      {
        flashOffState = true;
        for (let i=0; i < indFlashBytes.length; i++)
        {
          //invert the bits
          flashMaskBytes[i] = ~parseInt(indFlashBytes[i], 16);
        }
      }
      else
      {
        flashOffState = false;
        for (let i=0; i < indFlashBytes.length; i++)
        {
          //invert the bits
          flashMaskBytes[i] = 0xFF;
        }
      }

      flashTimeCntr = 0;
    }
    else
    {
      flashTimeCntr++;
    }

    //only process further and post a message if there has been a change since the last request.
    //Also, we only need to process after receiving both the data and flash data.
    //if (!reqIndFlag && ((indicationData !== lastIndicationData) || (flashData !== lastFlashData)))
    if (!reqIndFlag)
    {
      //apply flash mask to the indication bytes
      for (let i=0; i < indicationBytes.length; i++)
      {
        indicationBytes[i] = (parseInt(indicationBytes[i],16) & flashMaskBytes[i]).toString(16);
      }

      //console.log(indicationBytes);

      //unpack the bits
      for (let i=0; i < indicationBits.length; i++)
      {
        indicationBits[i] = GetBitState(i, indicationBytes);
      }
      for (let i=0; i < indicationMaskBits.length; i++)
      {
        indicationMaskBits[i] = GetBitState(i, indicationMaskBytes);
      }

      //console.log(indicationBits);

      postMessage({indication:indicationBits, indicationMask:indicationMaskBits, link:linkHealth});

      lastIndicationData = indicationData;
      lastFlashData = flashData;
    }
  }

  setTimeout("RequestIndications()", requestInterval);
}

//This function determines the bit state by parsing the specified array
function GetBitState(bit,byteArray)
{
  //start with LSB
  let bitmask = 1;

  //find the byte number
  let byteNbr = Math.floor(bit/8);

  //Find the remainder value and shift the bitmask that many times
  for (let i=0; i < bit % 8; i++)
  {
    //shift mask to left one bit
    bitmask = bitmask * 2;
  }

  //Convert the hex string to an integer and then AND it with
  //the bitmask to determine the bit state.
  if (parseInt(byteArray[byteNbr],16) & bitmask)
  {
    return 1;
  }
  else
  {
    return 0;
  }
}

function qnetSendControl(bit, state)
{
  let controlData = bit + "," + state;
  let Http = new XMLHttpRequest();
  Http.open("POST", urlRoot + "controls/bit/", true);
  Http.setRequestHeader("Content-type", "text/plain");
  Http.send(controlData);
  Http.onreadystatechange= function(){
    //not currently used
  };
}

function qnetSendAllControls(controlBits)
{
  //pack bits into comma separated control bytes
  let controlBytes = new Array(); //array to store the control bytes
  let controlData; //control string to send
  let byteValue = 0;

  //start with LSB
  let bitmask = 1;
  let byteNbr = 0;

  //Find set bits and build the controlBytes array.
  for (let i=0; i < controlBits.length; i++)
  {
    if (controlBits[i])
    {
      byteValue = byteValue | bitmask;
    }

    //shift mask to left one bit
    bitmask = bitmask * 2;

    if (256 === bitmask)
    {
      //done with that byte, save it and prepare for next byte
      controlBytes.push(byteValue);
      bitmask = 1;
      byteNbr++;
      byteValue = 0;
    }
  }

  //create control data string (convert to hex)
  controlData = controlBytes[0].toString(16);

  for (let i=1; i < controlBytes.length; i++)
  {
    controlData = controlData + "," + controlBytes[i].toString(16);
  }

  // console.log(controlData);

  let Http = new XMLHttpRequest();
  Http.open("POST", urlRoot + "controls/all/", true);
  Http.setRequestHeader("Content-type", "text/plain");
  Http.send(controlData);
  Http.onreadystatechange= function(){
    //not currently used
  };
}